import math
import os
import glob
from typing import List, Tuple, Dict
from torch.utils.data import Dataset, DataLoader
import torch
import torchvision
import torchvision.transforms as T
from torchvision.io import read_image


class Imagedataset(Dataset):
    def __init__(self, path: str):
        self.files = []
        for filename in glob.iglob(f'{path}/**', recursive=True):
            if os.path.isfile(filename):
                self.files.append(filename)
        self.n_samples = len(self.files)
        self.transforms = torch.nn.Sequential(
            T.Resize((64, 64)),
            T.ConvertImageDtype(torch.float),
            T.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
        )

    def __getitem__(self, index):
        raw_image = read_image(self.files[index])
        transformed_image = self.transforms(raw_image)
        return transformed_image

    def __len__(self):
        return self.n_samples


def __generate_embeddings(loader: DataLoader, model=None):
    """
    Generates embeddings from images by simply doing a forward pass.
    This seems to work unreasonably well.
    The embedding is simply the last layer of the model. By default using
    models trained on imagenet this will be dim=1000. Notice that using
    normal CNN trained on image classification might not be SOTA for
    this, but it seems to work darn well!

    By default it uses a resnet18, but the user can also specify a custom model.
    Function will use the output of specified custom model as the embedding.

    This is also the function that takes over 99 % of the time.
    :param loader:
    :param model:
    :return:
    """
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    if model is None:
        model = torchvision.models.resnet18(pretrained=True).to(device)
    model.eval()
    # returns embeddings in shape (n, 1000)
    embeddings = []
    for data in loader:
        embedding = model(data.to(device))
        embeddings.append(embedding.flatten(start_dim=1).detach().cpu())
    return torch.cat(embeddings)


def __knn(emb_train, emb_test, K: int, batch_size: int, threshold=30):
    """
    KNN that supports GPU. VERY fast.

    :param emb_train:
    :param emb_test:
    :param K:
    :param batch_size:
    :param threshold:
    :param print_dupes:
    :return:
    """
    for binx in range(math.ceil(len(emb_test) / batch_size)):
        # Slice the test set
        batch_test_emb = emb_test[binx * batch_size: binx * batch_size + batch_size, :]
        # Compute dist between test embeddings and train embeddings
        D = torch.cdist(batch_test_emb, emb_train)
        # Get distances and indexes to top k closest vectors
        dist, idx = D.topk(k=K, dim=-1, largest=False)
        duplicates = []
        for i, row in enumerate(idx):
            for j, elem in enumerate(row):
                if float(dist[i][j]) < threshold:
                    duplicates.append((binx * batch_size + int(elem), i))
        return duplicates


def duplicate_images(train_loader: DataLoader, test_loader: DataLoader,
                     K:int, batch_size:int, model=None) -> List[Tuple[int, int]]:
    """
    Returns indexes for images that are very similar or identical.

    Generates embeddings by doing a forward pass with a pretrained resnet18 (by default).
    Alternatively you can specify a custom model for generating embeddings.
    The embeddings are then compared with KNN to find similar images.

    :param train_loader:
    :param test_loader:
    :param K:
    :param batch_size:
    :param model:
    :param print_dupes:
    :return:
    """
    emb_train = __generate_embeddings(train_loader, model)
    emb_test = __generate_embeddings(test_loader, model)
    duplicates = __knn(emb_train, emb_test, K, batch_size)
    return duplicates


def duplicate_images_dir(dir: str, K: int, batch_size: int,
                         model=None, num_workers: int = 0,
                         print_dupes: bool = False) -> List[Tuple[str, str]]:
    """
    Returns file paths of images that are near duplicates or identical.
    Example output [(img1.jpg, img2.jpg) ...]

    :param dir:
    :param K:
    :param batch_size:
    :param recursive:
    :param model:
    :param num_workers:
    :param print_dupes:
    :return:
    """
    duplicate_files = []
    dataset = Imagedataset(dir)
    files = dataset.files
    data_loader = torch.utils.data.DataLoader(dataset, batch_size=batch_size,
                                              shuffle=False, num_workers=num_workers)
    embeddings = __generate_embeddings(data_loader, model)
    duplicates = __knn(embeddings, embeddings, K, batch_size)
    for dupe in duplicates:
        if dupe[0] != dupe[1]:
            image1 = files[dupe[0]]
            image2 = files[dupe[1]]
            if print_dupes:
                print(image1, "-->", image2)
            duplicate_files.append((image1, image2))
    return duplicate_files



from torch.utils.data import DataLoader, Dataset
from pathlib import Path
import unittest
import sys
import torchvision
import torchvision.transforms as transforms

#### needed for testing #####
this_path = str(Path(__file__).parent.parent.parent)
sys.path.append(this_path)
#############################

import did_it_spill


class IntegerDatasetForTesting(Dataset):

    def __init__(self, start=0, end=10):
        assert start >= 0
        assert end >= 0
        assert end >= start

        self.x_data = list(range(start, start + end))
        self.y_data = list(range(start, start + end))

    def __len__(self):
        return len(self.x_data)

    def __getitem__(self, index):
        return self.x_data[index], self.y_data[index]


class CharacterDatasetForTesting(Dataset):

    data = list()

    def __init__(self, size=1):
        assert size > 0
        self.__generate_data(size=size)

    def __generate_data(self, size):
        resulting_data_list = list()

        for i in range(size):
            i = i % 26
            char_i = i + 97 # i=0 this will be an "a"
            amount_of_char = i // 26 # when size is greater than 26 this dataset will no longer be unique
            string_char_representation = str(chr(char_i))*(amount_of_char+1)
            resulting_data_list.append(string_char_representation)
        self.data = resulting_data_list

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        return self.data[index], 1


class TestMain(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        super(TestMain, cls).setUpClass()
        # taken from pytorch tutorial for classification: https://pytorch.org/tutorials/beginner/blitz/cifar10_tutorial.html
        transform = transforms.ToTensor()
        cls.cifar10_train = torchvision.datasets.CIFAR10(root="./data", train=True, download=True, transform=transform)
        cls.cifar10_test = torchvision.datasets.CIFAR10(root="./data", train=False, download=False, transform=transform)

    def test_overlap_with_batchsize_1(self):
        self.__test_overlap()

    def test_no_overlap_with_batchsize_1(self):
        self.__test_no_overlap()

    def test_overlap_with_batchsize_2(self):
        self.__test_overlap(batch_size=2)

    def test_no_overlap_with_batchsize_2(self):
        self.__test_no_overlap(batch_size=2)

    def test_overlap_stringdata_with_batchsize_1(self):
        train_dataset = CharacterDatasetForTesting(size=10)
        test_dataset = CharacterDatasetForTesting(size=2)

        train_dataloader = self.__create_dataloader(train_dataset, batch_size=1, shuffle=False)
        test_dataloader = self.__create_dataloader(test_dataset, batch_size=1, shuffle=False)

        # raised since the dataloader will not pack string or chars into a tensor
        self.assertRaises(AttributeError, did_it_spill.check_spill, train_dataloader, test_dataloader)

    def test_cifar10_no_spills(self):
        train_dataloader = self.__create_dataloader(self.cifar10_train)
        test_dataloader = self.__create_dataloader(self.cifar10_test)

        spills = did_it_spill.check_spill(train_dataloader, test_dataloader)

        expected_indexes_in_overlap = set()

        self.__check_expected_values(spills, expected_indexes_in_overlap)

    def test_cifar10_full_spills(self):
        test_dataloader = self.__create_dataloader(self.cifar10_test)

        spills = did_it_spill.check_spill(test_dataloader, test_dataloader)

        #since we compare the identical dataloaders all entries should overlap
        expected_indexes_in_overlap = set(range(len(test_dataloader)))

        self.__check_expected_values(spills, expected_indexes_in_overlap)

    def __test_overlap(self, batch_size=1, shuffle=False):
        train_dataset = IntegerDatasetForTesting(start=0, end=10)
        test_dataset = IntegerDatasetForTesting(start=0, end=5)

        train_dataloader = self.__create_dataloader(train_dataset, batch_size=batch_size, shuffle=shuffle)
        test_dataloader = self.__create_dataloader(test_dataset, batch_size=batch_size, shuffle=shuffle)

        expected_indexes_in_overlap = {0, 1, 2, 3, 4}
        spills = did_it_spill.check_spill(train_dataloader, test_dataloader)

        self.__check_expected_values(spills, expected_indexes_in_overlap)

    def __test_no_overlap(self, batch_size=1, shuffle=False):
        train_dataset = IntegerDatasetForTesting(start=0, end=5)
        test_dataset = IntegerDatasetForTesting(start=10, end=15)

        train_dataloader = self.__create_dataloader(train_dataset, batch_size=batch_size, shuffle=shuffle)
        test_dataloader = self.__create_dataloader(test_dataset, batch_size=batch_size, shuffle=shuffle)

        expected_indexes_in_overlap = set()
        spills = did_it_spill.check_spill(train_dataloader, test_dataloader)

        self.__check_expected_values(spills, expected_indexes_in_overlap)

    @staticmethod
    def __create_dataloader(dataset, batch_size=1, shuffle=False):
        return DataLoader(dataset=dataset, shuffle=shuffle, batch_size=batch_size)

    def __check_expected_values(self, spills, expected_indexes_in_overlap):
        seen = dict()
        for spill in spills:
            self.assertTrue(spill[0] in expected_indexes_in_overlap)
            self.assertTrue(spill[1] in expected_indexes_in_overlap)
            self.assertTrue(spill[0] == spill[1])
            seen[spill[0]] = True

        self.assertTrue(len(seen) == len(spills))





import numpy as np
from typing import List, Tuple, Dict
from numpy.typing import NDArray
import hashlib

# Not done yet

def __get_hash_dict(data: NDArray) -> Dict[str, int]:
    index = 0
    hash_dict = {}
    for sample in data:
        # Hash the current sample
        this_hash = hashlib.sha1(sample).hexdigest()
        # Key = hash of our sample, val = where in the dataset that hash was
        hash_dict[this_hash] = index
        index += 1
    return hash_dict


def check_spill(train: NDArray, test: NDArray) -> List[Tuple[int, int]]:
    test_spills = []
    hashes_loader1 = __get_hash_dict(train)
    hashes_loader2 = __get_hash_dict(test)
    # find spilled samples
    for hsh in hashes_loader1.keys():
        if hsh in hashes_loader2.keys():
            test_spills.append((hashes_loader1[hsh], hashes_loader2[hsh]))
    return test_spills



from typing import List, Tuple, Dict
import numpy as np
import hashlib
from torch.utils.data import DataLoader, Dataset


def __get_hash_dict(loader: DataLoader) -> Dict[str, int]:
    index = 0
    hash_dict = {}
    # For each batch
    for data, _ in loader:
        data = data.detach().numpy()
        # For each sample in the batch
        for sample in data:
            # Hash the current sample
            this_hash = hashlib.sha1(sample).hexdigest()
            # Key = hash of our sample, val = where in the dataset that hash was
            hash_dict[this_hash] = index
            index += 1
    return hash_dict


def check_spill(loader1: DataLoader, loader2: DataLoader) -> List[Tuple[int, int]]:
    test_spills = []
    hashes_loader1 = __get_hash_dict(loader1)
    hashes_loader2 = __get_hash_dict(loader2)
    # find spilled samples
    for hsh in hashes_loader1.keys():
        if hsh in hashes_loader2.keys():
            test_spills.append((hashes_loader1[hsh], hashes_loader2[hsh]))
    return test_spills


def get_spilled_samples(spills: List, train_dataset: Dataset):
    """
    Returns the actual data that was spilled. Notice that it
    returns everything that the __getitem__ returns ie. data and labels
    and potentially other stuff. This is done to be more
    general, not just work with datasets that return: (data, label),
    but also for datasets with (data, label, third_thing) or similar.

    Notice that the function only takes in one dataset but spill
    is a tuple with indexes for two datasets (the other is ignored).
    :param spills:
    :param train_dataset:
    :return: spilled_samples:
    """
    spilled_samples = []
    for spill in spills:
        spill_inx = spill[0]
        spilled_samples.append(train_dataset.__getitem__(spill_inx))
    return spilled_samples
